// Copyright 2018-2023 Xanadu Quantum Technologies Inc.

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at

//     http://www.apache.org/licenses/LICENSE-2.0

// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
#include "GateIndices.hpp"
#include "ConstantUtil.hpp"
#include "Util.hpp" // exp2, maxDecimalForQubit

namespace Pennylane::LightningQubit::Gates {
auto getIndicesAfterExclusion(const std::vector<std::size_t> &indicesToExclude,
                              std::size_t num_qubits)
    -> std::vector<std::size_t> {
    std::set<std::size_t> indices;
    for (std::size_t i = 0; i < num_qubits; i++) {
        indices.emplace(i);
    }
    for (const std::size_t &excludedIndex : indicesToExclude) {
        indices.erase(excludedIndex);
    }
    return {indices.begin(), indices.end()};
}

auto generateBitPatterns(const std::vector<std::size_t> &qubitIndices,
                         std::size_t num_qubits) -> std::vector<std::size_t> {
    std::vector<std::size_t> indices;
    indices.reserve(Pennylane::Util::exp2(qubitIndices.size()));
    indices.emplace_back(0);

    // NOLINTNEXTLINE(modernize-loop-convert)
    for (auto index_it = qubitIndices.rbegin(); index_it != qubitIndices.rend();
         index_it++) {
        const std::size_t value =
            Pennylane::Util::maxDecimalForQubit(*index_it, num_qubits);
        const std::size_t currentSize = indices.size();
        for (std::size_t j = 0; j < currentSize; j++) {
            indices.emplace_back(indices[j] + value);
        }
    }
    return indices;
}

/**
 * @brief Introduce quantum controls in indices generated by
 * generateBitPatterns.
 *
 * @param indices Indices for the operation.
 * @param num_qubits Number of qubits in register.
 * @param controlled_wires Control wires.
 * @param controlled_values Control values (false or true).
 */
void controlBitPatterns(std::vector<std::size_t> &indices,
                        const std::size_t num_qubits,
                        const std::vector<std::size_t> &controlled_wires,
                        const std::vector<bool> &controlled_values) {
    constexpr std::size_t one{1U};
    if (controlled_wires.empty()) {
        return;
    }
    std::vector<std::size_t> controlled_values_i(controlled_values.size());
    std::transform(controlled_values.begin(), controlled_values.end(),
                   controlled_values_i.begin(),
                   [](const bool v) { return static_cast<std::size_t>(v); });
    std::for_each(
        indices.begin(), indices.end(),
        [num_qubits, &controlled_wires, &controlled_values_i](std::size_t &i) {
            for (std::size_t k = 0; k < controlled_wires.size(); k++) {
                const std::size_t rev_wire =
                    (num_qubits - 1) - controlled_wires[k];
                const std::size_t value = controlled_values_i[k];
                i = (i & ~(one << rev_wire)) | (value << rev_wire);
            }
        });
}
} // namespace Pennylane::LightningQubit::Gates
